#ifndef EXPR_H
#define EXPR_H

#include "adout.h"	/* for typedef 'Alloc' */
#include "lex.h"
#include "symbol.h"
#include "tokens.h"
#include "typedef.h"

/*{{{  #defines*/
/*{{{  'ep->token' values  (ND_...)*/
/*{{{  operators (ND_..._OP)*/
/*-------------------------------------------------------------
Operator nodes
--------------
N.B.
(1) We don't use the highest bit in 'ep->token', so it can have
the signed type 'int32' rather than the unsigned type 'bits32',
and still always has a positive value.
This means that we can take the difference of two token values
and get a signed result, which is useful for 'ordering' expression
nodes, as done in 'compare_nodes (ep1,ep2)'.

(2) A problem:  'ND_DEFINED_OP' should have 2 precedence levels,
depending on whether it's unary or binary!
(N.B. this isn't so for '+' and '-').

(3) This assumes that 'int32' is 'long'.
-------------------------------------------------------------*/

#define  ND_NON_ASSOC		0L
#define  ND_LEFT_ASSOC		0x80000L
#define  ND_RIGHT_ASSOC		0x100000L

#define  ND_LEFT_BRANCH		ND_LEFT_ASSOC
#define  ND_RIGHT_BRANCH	ND_RIGHT_ASSOC

#define  ND_PREC_1		0x200000L
#define  ND_PREC_2		0x400000L
#define  ND_PREC_3		0x800000L
#define  ND_PREC_4		0x1000000L
#define  ND_PREC_5		0x2000000L
#define  ND_PREC_6		0x4000000L
#define  ND_PREC_7		0x8000000L
#define  ND_PREC_8		0x10000000L
#define  ND_PREC_9		0x20000000L
#define  ND_PREC_10		0x40000000L

#define  ND_PREC_MASK		0x7FE00000L	/* (ND_PREC_1 | ... | _10) */
#define  ND_OPER_MASK		ND_PREC_MASK	/* i.e. is it an operator? */
				
#define  ND_DEFINED_OP		( 1 | ND_LEFT_ASSOC  | ND_PREC_1)
#define  ND_EQV_OP		( 2 | ND_LEFT_ASSOC  | ND_PREC_2)
#define  ND_NEQV_OP		( 3 | ND_LEFT_ASSOC  | ND_PREC_2)
#define  ND_OR_OP		( 4 | ND_LEFT_ASSOC  | ND_PREC_3)
#define  ND_AND_OP		( 5 | ND_LEFT_ASSOC  | ND_PREC_4)
#define  ND_NOT_OP		( 6 | ND_LEFT_ASSOC  | ND_PREC_5)
#define  ND_REL_OP		( 7 | ND_NON_ASSOC   | ND_PREC_6)
#define  ND_CONCAT_OP		( 8 | ND_LEFT_ASSOC  | ND_PREC_7)
#define  ND_PLUS_OP		( 9 | ND_LEFT_ASSOC  | ND_PREC_8)
#define  ND_U_PLUS_OP		(10 | ND_LEFT_ASSOC  | ND_PREC_8)
#define  ND_MINUS_OP		(11 | ND_LEFT_ASSOC  | ND_PREC_8)
#define  ND_U_MINUS_OP		(12 | ND_LEFT_ASSOC  | ND_PREC_8)
#define  ND_TIMES_OP		(13 | ND_LEFT_ASSOC  | ND_PREC_9)
#define  ND_DIV_OP		(14 | ND_LEFT_ASSOC  | ND_PREC_9)
#define  ND_POWER_OP		(15 | ND_RIGHT_ASSOC | ND_PREC_10)
/*}}}*/
/*{{{  primaries*/
/*----------------------*/
/* Nodes for primaries.	*/
/*----------------------*/
/* a literal constant */

#define  ND_CHAR_LIT_CONST	((int32) 16)
#define  ND_COMPLEX_LIT_CONST	((int32) 17)
#define  ND_DP_LIT_CONST	((int32) 18)
#define  ND_INT_LIT_CONST	((int32) 19)
#define  ND_LOGICAL_LIT_CONST	((int32) 20)
#define  ND_REAL_LIT_CONST	((int32) 21)

/* an ID not followed by (..)  (except possibly a substring designator, (l:u))*/

#define  ND_SIMPLE_SCALAR	((int32) 22)
#define  ND_TEMP_SCALAR		((int32) 23)
#define  ND_WHOLE_ARRAY		((int32) 24)
#define  ND_PROC_ACTUAL_ARG	((int32) 25)

/* an ID followed by (..), i.e. by an arg, subscript or component list */

#define  ND_ARRAY_SECT		((int32) 26)	/* ...or array element */
#define  ND_TEMP_ARRAY		((int32) 27)

#define  ND_FUNC		((int32) 28)
#define  ND_SUBR		((int32) 29)
#define  ND_STRUCT_CONSTR	((int32) 30)

/* an argument, subscript, component or substring node */

#define  ND_EXPRLIST		((int32) 31)

#define  ND_SUBSCR_DOUBLET	((int32) 32)	/* ...or substring range */
#define  ND_SUBSCR_TRIPLET	((int32) 33)
#define  ND_ALT_RETURN_LABEL	((int32) 34)

/* an internally generated expr tree, where the primary's name
   is stored in a string rather than given by an 'sp' pointer  --*/

#define  ND_INTERN_OBJECT	((int32) 35)
#define  ND_INTERN_FUNC		((int32) 36)

/* '*' and error nodes */

#define  ND_STAR		((int32) 37)
#define  ND_ERROR		((int32) 38)

#if 0
/*{{{  <NOTE> primaries' 'val', 'left' & 'right' fields, & constructor functions*/
-------------------------------------------------------------------
'val', 'left' and 'right' fields of primary expr nodes, and the
functions that create the nodes  (i.e. their constructor functions).
-------------------------------------------------------------------
			e->val                           constructor  
			member    e->left     e->right    function          
			-----     -------     --------   -----------
-- a literal constant

ND_CHAR_LIT_CONST	strval     [substr]   -          str_lit_primary
ND_COMPLEX_LIT_CONST	strval     -          -               "     "
ND_DP_LIT_CONST		strval     -          -               "     "
ND_INT_LIT_CONST	intval     -          -          int_lit_primary
ND_LOGICAL_LIT_CONST	strval     -          -          str_lit_primary
ND_REAL_LIT_CONST	strval     -          -               "     "

-- an ID not followed by (..)  (except perhaps a substring designator, (l:u))

ND_SIMPLE_SCALAR	sp         [substr]   -          named_data_ / stmt_var_primary
ND_WHOLE_ARRAY		sp         -          -          named_data_primary
ND_PROC_ACTUAL_ARG	sp         -          -          proc_name_actual

-- an ID followed by (..), i.e. by an arg, subscript or component list

ND_ARRAY_SECT		sp         [substr]   [exprlist]  named_data_primary
ND_TEMP_ARRAY		allocp     -          [var]       alloc_var / _mapvec,
			                                    replace_expr_by_temp
ND_TEMP_SCALAR		allocp     -          [var]       alloc_var,
			                                    replace_expr_by_temp,
			                                    save_scalar_in_temp

ND_FUNC			sp         -          [exprlist]  proc_ref_primary
ND_SUBR			sp         -          [exprlist]  proc_ref_primary
ND_STRUCT_CONSTR	sp         -          [exprlist]  struct_constr_primary

-- an argument, subscript, component or substring node

ND_EXPRLIST		[sp]       expr       [exprlist]  exprlist_item, keyword_arg

ND_SUBSCR_DOUBLET	-          [l]        [u]        subscr_triplet
ND_SUBSCR_TRIPLET	-          doublet    s          subscr_triplet
ND_ALT_RETURN_LABEL	intval     -          -          alt_return_label

-- an internally generated expr tree, where the primary's name
   is stored in a string rather than given by an 'sp' pointer

ND_INTERN_OBJECT	strval     -          -          name_ / map_primary
ND_INTERN_FUNC		strval     -          -          intern_func_ref

-- '*' and error nodes

ND_STAR			-          -          -          init_primaries
ND_ERROR		-          -          -          init_primaries

/*}}}*/
#endif
/*}}}*/
/*}}}*/
/*{{{  'ep->type'  flags   (EX_...)*/
/*--------------------------------------------------------------*/
/* 'ep->type' flags (EX_...)					*/
/*								*/
/* N.B.:							*/
/* (1) This assumes that 'bits32' is 'long' (signed or unsigend)*/
/* (2) THE DATATYPE FLAGS ARE THE SAME AS THOSE FOR 'sp->type'	*/
/*     (defined in "symbol.h").					*/
/*--------------------------------------------------------------*/
/*--------------*/
/* single flags */
/*--------------*/
/*-- data types (**SAME AS FOR 'sp->type'**) --*/

#define  EX_KIND_MASK		TY_KIND_MASK	/* 0x3FL */

#define  EX_INTEGER		TY_INTEGER	/* 0x40L */
#define  EX_REAL		TY_REAL		/* 0x80L */
#define  EX_DP			TY_DP		/* 0x100L */
#define  EX_COMPLEX		TY_COMPLEX	/* 0x200L */
#define  EX_LOGICAL		TY_LOGICAL	/* 0x400L */
#define  EX_CHARACTER		TY_CHARACTER	/* 0x800L */
#define  EX_DERIVED		TY_DERIVED	/* 0x1000L */
#define  EX_NONE		TY_NONE		/* 0x2000L */

/*-- attributes --*/

#define  EX_CONST		0x4000L
#define  EX_SCALAR		0x8000L

#define  EX_INT_LIT_CONST	0x10000L	/* int-lit-const */
#define  EX_INT_SUM		0x20000L	/* int-expr + int-expr */
#define  EX_INT_SUM_CONST	0x40000L	/* int-expr + int-lit-const */
#define  EX_INT_PROD		0x80000L	/* int-expr * int-expr */
#define  EX_INT_PROD_CONST	0x100000L	/* int-expr * int-lit-const */
#define  EX_INT_RATIO		0x200000L	/* int-expr / int-expr */

/*-----------------*/
/* composite flags */
/*-----------------*/
/*-- datatype masks (** SAME AS FOR 'sp->type'**) --*/

#define  EX_BASIC_TYPE		TY_BASIC_TYPE	/* type *without* KIND val */
#define  EX_TYPE		TY_TYPE		/* type *with* KIND val */
#define  EX_INT_REAL_DP		TY_INT_REAL_DP	/* integer / real / dp */
#define  EX_NUMERIC		TY_NUMERIC	/* int/ real/ dp/ complex */
#define  EX_INT_CHAR_LOGICAL	(TY_INTEGER | TY_CHARACTER | TY_LOGICAL)
						/* ...for CASE exprs */
/*-- other masks --*/

#define  EX_INT_EXPR_TYPE	(EX_INT_LIT_CONST |\
				 EX_INT_SUM | EX_INT_SUM_CONST |\
				 EX_INT_PROD | EX_INT_PROD_CONST |\
				 EX_INT_RATIO)

#define  EX_INT_PLUS		(EX_INT_SUM | EX_INT_SUM_CONST)
#define  EX_INT_TIMES		(EX_INT_PROD | EX_INT_PROD_CONST)
/*}}}*/
/*{{{  'sectp->attrs' flags  (SE_...)*/
/*--------------------------------------*/
/* 'sectp->attrs' flags (SE_...)	*/
/*--------------------------------------*/
#define  SE_RANK_MASK		0x7	/* field for 'sect_rank' */
#define  SE_LOCALISED_SECT	0x8	/* set if 'sectp' parameters are
					   'localised' */
/*}}}*/

/*-----------------------------------*/
/* subscripts of 'seg_ep []' vector! */
/*-----------------------------------*/
#define  ETA   0
#define  SEGL  1
#define  SEGN  2

/*--------------------------------------------------------------*/
/* Flags used as actual parameters to 'delete_expr()', where	*/
/* they're associated with the formal parameter 'decr_nlinks'.	*/
/*--------------------------------------------------------------*/
#define  DECR_NLINKS	TRUE
#define  LEAVE_NLINKS	FALSE

#define  NULL_Expr	((Expr *) NULL)
#define  NULL_Expr_p	((Expr **) NULL)
#define  NULL_Sect	((Sect *) NULL)
/*}}}*/
/*{{{  typedefs*/
typedef union {
	struct Alloc *allocp;	/* pointer to temporary array (in Alloc stack!) */
	struct Symbol *sp;	/* pointer to symbol table entry */
	int  intval;		/* int value for INT_LIT_CONSTs */
	char *strval;		/* string value pointer */
} Exprval;


/*--------------------------------------------------------------*/
/* 'Sect':  a structure for regular section / array element	*/
/* parameters.							*/
/*--------------------------------------------------------------*/
typedef struct Sect {
	short   nlinks;		/* # links pointing to this node */
	short   attrs;
	struct Expr  *mapvec_ep;	/* expr for section's 'mapvec' */
	/*------------------------------------------------------*/
	/* The following are vectors of length 'sect_rank'	*/
	/* (== array_rank, as given by 'ep->mapp->attrs' and	*/
	/* 'ep->sectp->attrs').					*/
	/*							*/
	/* N.B.   Logically there are 3 of them.  However, in	*/
	/* fact they're allocated as a single vector, '*l', of	*/
	/* length [3 *sect_rank] containing (l[], n[], s[]).	*/
	/*------------------------------------------------------*/
	struct Expr  **l,  **n,  **s;
} Sect;


/*------------------------------------------*/
/* 'Expr': structure for an expression node */
/*------------------------------------------*/
typedef struct Expr {
	int32         token;	/* ND_... */
	Exprval       val;	/* Value of the object at this node:
				     sp      (eg for ND_SCALAR);
				     intval  (eg for ND_INT_LIT_CONST);
				     strval  (eg for ND_REL_OP).	*/
	bits32        type;	/* type | EX_SCALAR | EX_CONST... etc.
				     LATER ON, THIS COULD CONTAIN ARRAY INFO,
				     ARGUMENT INFO, ETC?? */
	short         nlinks;	/* # links pointing to this node */
	struct Map   *mapp;
	Sect         *sectp;
	struct Expr  *left;
	struct Expr  *right;
} Expr;
/*}}}*/
/*{{{  externs*/
bool     exprs_equal (),  expr_mustbe (),  occurs_in_expr (),
	 shapes_equal ();

Expr	*abs_func (),   *alloc_mapvec (),   *alloc_var (),
	*alt_return_label (),   *array_extent_func (),
	*array_upper_bound (),   *ceiling_div (),
	*change_int_lit_primary (),   *check_and_transform_format_expr (),
	*copy_node (),   *copy_ops (),   *defined_expr (),
	*dummy_map_primary (),   *expr (),
	*exprlist_item (),   *extract_minus_signs (),   *find_array_term (),
	*int_lit_primary (),   *keyword_arg (),   *map_primary (),
	*mod_func (),    *name_primary (),   *named_data_primary (),
	*negate_addop_expr (),   *no_errmsg_div (),   *prefix_binary_op (),
	*proc_name_actual (),   *proc_coord_primary (),
	*proc_ref_primary (),   *rel_expr (),   *save_scalar_in_temp (),
	*sect_extent_func (),   *simplify_expr (),   *stmt_var_primary (),
	*str_lit_primary (),   *struct_constr_primary (),
	*subscr_triplet (),   *vec_len ();

#ifdef  __STDC__
Expr  *intern_func_ref (char  *func_name,  int  nargs,  ...);
#else
Expr  *intern_func_ref ();
#endif

extern bool   spec_expr;
extern Expr  *error_primary,  *int_lit_0,  *int_lit_1,  *int_lit_2,
             *int_lit_m1,  *int_lit_m2,  *star_primary;
/*}}}*/
/*{{{  macros*/
/*--------------------------------------------------------------*/
/* 'int_val(ep)' returns the value of 'ep', which must be an	*/
/* integer literal constant exprn.  The value is held in	*/
/* 'ep->val.intval', even if it's -ve, in which case 'ep' is a	*/
/* 'ND_MINUS_OP' node whose right operand is a			*/
/* 'ND_INT_LIT_CONST' node.					*/
/*--------------------------------------------------------------*/
#define  int_val(ep)		((ep)->val.intval)


/*--------------------------------------------------------------*/
/* assign_ptr (to_p, fr_p)					*/
/* -----------------------					*/
/* 'to_p' and 'fr_p' are pointers to the same type of structure,*/
/* which must have a member 'nlinks' that counts the number of	*/
/* pointers pointing to it.  Objects of this type are typically	*/
/* (or always) malloc'ed, and 'nlinks' is required to determine	*/
/* when it can be freed.  (Examples are Expr*, Array*, Distr*).	*/
/*	assign_ptr (to_p,fr_p)					*/
/* sets 'to_p = fr_p' and also increments 'fr_p->nlinks'.	*/
/*								*/
/* N.B. The principles of this 'garbage collection'  scheme	*/
/* are described at the start of 'expr.c'			*/
/*--------------------------------------------------------------*/
#define  assign_ptr(to_p,fr_p)	(((to_p) = (fr_p))->nlinks++)
#define  attach_expr		assign_ptr
#define  attach_array		assign_ptr
#define  attach_align		assign_ptr
#define  attach_distr		assign_ptr
#define  attach_map		assign_ptr
#define  attach_sect		assign_ptr

#define  protect_expr(ep)	((ep)->nlinks++)
#define  unprotect_expr(ep)	(delete_expr ((ep), DECR_NLINKS))

/*--------------------------------------------------------------*/
/* prod_signs (op1, op2)					*/
/* ---------------------					*/
/* 'op1', 'op2' are 'add-op' tokens, i.e. they have the values	*/
/* ND_PLUS_OP, ND_U_PLUS_OP, ND_MINUS_OP or ND_U_MINUS_OP.	*/
/* The result is the product of the signs 'op1' and 'op2', and	*/
/* has the value ND_PLUS_OP or ND_MINUS_OP.			*/
/*								*/
/* N.B. It's assumed that this is only used to derive a		*/
/* *binary* operator token.					*/
/*--------------------------------------------------------------*/
#define  prod_signs(op1,op2)						\
		((((op1) == ND_PLUS_OP || (op1) == ND_U_PLUS_OP) !=	\
		  ((op2) == ND_PLUS_OP || (op2) == ND_U_PLUS_OP)) ?	\
		 ND_MINUS_OP : ND_PLUS_OP)

#define  change_sign(op)						\
		(((op) == ND_PLUS_OP || (op) == ND_U_PLUS_OP) ?		\
		 ND_MINUS_OP : ND_PLUS_OP)
/*}}}*/

#endif
